# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Pyramid-Helpers common views """

import hashlib

from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPInternalServerError
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.security import forget
from pyramid.security import remember
from pyramid.view import forbidden_view_config
from pyramid.view import notfound_view_config
from pyramid.view import view_config

from pyramid_helpers.api_doc import api_doc as api_doc_
from pyramid_helpers.auth import check_credentials
from pyramid_helpers.forms import validate
from pyramid_helpers.forms.auth import LoginForm
from pyramid_helpers.forms.validators import ValidatorsForm
from pyramid_helpers.i18n import N_
from pyramid_helpers.utils import get_settings


# HTTPExceptions' titles
N_('Bad Request')
N_('Forbidden')
N_('Not Found')

N_('Access was denied to this resource.')
N_('The resource could not be found.')


@view_config(route_name='api-doc', renderer='')
def api_doc(request):
    """ The API doc view """

    _ = request.translate

    renderer_input = dict(
        breadcrumb=[(_('API documentation'), None)],
        subtitle=None,
        title=_('API documentation'),
    )

    return api_doc_(request, renderer_input, 'pyramid_helpers')


@forbidden_view_config(renderer='/errors.mako')
def http_403(request):
    """ HTTP 403 error view """

    _ = request.translate

    if request.authenticated_user is None:
        params = get_settings(request, 'auth', 'auth')
        if params.get('policies') == ['basic']:
            # Issuing a challenge
            response = HTTPUnauthorized()
            response.headers.update(forget(request))
            return response

        return HTTPFound(location=request.route_path('login'))

    request.response.status_int = 403

    return dict(
        title=_(request.exception.title),
        subtitle='',
    )


@notfound_view_config(renderer='/errors.mako')
def http_404(request):
    """ HTTP 404 error view """

    _ = request.translate

    request.response.status_int = 404

    return dict(
        title=_(request.exception.title),
        subtitle='',
    )


@view_config(route_name='index')
def index(request):
    """ Index view """
    return HTTPFound(location=request.route_path('articles.search'))


@view_config(route_name='login', renderer='/login.mako')
@validate('login_form', LoginForm)
def login(request):
    """ Login view """

    _ = request.translate
    form = request.forms['login_form']

    params = get_settings(request, 'auth', 'auth')
    if params.get('policies') == ['remote']:
        params = get_settings(request, 'auth', 'policy:remote')
        login_url = params.get('login_url')
        if login_url is None:
            raise HTTPInternalServerError(explanation=_('Authentication not available, please retry later.'))

        return HTTPFound(location=login_url)

    if request.method == 'POST':
        if not form.errors:
            username = form.result['username']
            password = form.result['password']
            came_from = form.result['came_from']

            if check_credentials(username, password, request):
                headers = remember(request, username)
                return HTTPFound(location=came_from, headers=headers)

            form.errors['username'] = _('Bad user or password')
    else:
        # Default data
        data = dict(came_from=request.referer if request.referer and request.referer != request.url else request.route_path('index'))
        form.set_data(data)

    return dict(
        title=_('Authentication'),
    )


@view_config(route_name='logout', renderer='/logout.mako')
def logout(request):
    """ Logout view """

    _ = request.translate
    session = request.session

    # Clear session
    session.delete()

    if request.authentication_policy == 'remote':
        params = get_settings(request, 'auth', 'policy:remote')
        logout_url = params.get('logout_url')
        if logout_url is None:
            # Display logout page
            return dict(
                title=_(u'Logged out'),
            )
    else:
        logout_url = request.route_path('index')

    headers = forget(request)
    return HTTPFound(location=logout_url, headers=headers)


@view_config(route_name='validators', renderer='/validators.mako')
@validate('validators_form', ValidatorsForm)
def validators(request):
    """ Validators view """

    _ = request.translate
    form = request.forms['validators_form']

    if request.method == 'POST':
        if not form.errors:
            qfile = form.result['upload_input']
            if qfile is not None:
                content = qfile.file.read()
                form.result['upload_input.name'] = qfile.filename
                form.result['upload_input.size'] = len(content)
                form.result['upload_input.md5sum'] = hashlib.md5(content).hexdigest()

    return dict(
        breadcrumb=[(_('Validators'), None)],
        errors=form.errors,
        result=form.result,
    )



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Radius client for Pyramid-Helpers """

from io import StringIO
import logging
import re
import threading
import time

from beaker.util import NoneType
from beaker.util import verify_rules

from pyrad.client import Client
from pyrad.client import Timeout
from pyrad.dictionary import Dictionary
from pyrad.packet import AccessAccept
from pyrad.packet import AccessRequest

from pyramid_helpers.auth import AuthenticationClient
from pyramid_helpers.utils import get_settings


SETTINGS_DEFAULTS = {
    'radius': {
        'enabled': False,
        'retries': 2,
        'timeout': 3.0,
    },
    r'server:\d+': {
        'enabled': True,
        'acctport': 1813,
        'authport': 1812,
        'coaport': 3799,
    },
}

SETTINGS_RULES = {
    r'attr:\d+': [
        ('name', (str, ), 'name must be a string designating valid attribute name'),
        ('type', (str, ), 'type must be a string designating valid attribute type'),
    ],
    'radius': [
        ('enabled', (bool, NoneType), 'enabled must be a boolean or an integer'),
        ('retries', (int, ), 'retries must be a string designating a valid integer'),
        ('timeout', (float, ), 'timeout must be a string designating a valid float'),
    ],
    r'server:\d+': [
        ('enabled', (bool, NoneType), 'enabled must be a boolean or an integer'),
        ('server', (str, ), 'server must be a string designating valid server'),
        ('secret', (str, ), 'secret must be a string designating a valid secret'),
        ('authport', (int, ), 'authentication port must be an integer'),
        ('acctport', (int, ), 'accounting port must be an integer'),
        ('coaport', (int, ), 'CoA port must be an integer'),
    ],
}


log = logging.getLogger(__name__)

client = None


class RadiusClient(AuthenticationClient):
    """ RADIUS authentication client """

    __auth_mode__ = 'radius'

    def __init__(self):
        """ RADIUS initialization """

        self.dictionary = None
        self.retries = None
        self.servers = []
        self.timeout = None

        self.__lock = threading.Lock()

    def __enter__(self):
        """ Acquire lock """

        self.__lock.acquire()

        return self

    def __exit__(self, type_, value, traceback):
        """ Release lock """

        self.__lock.release()

    def setup(self, *args, **kwargs):
        """
        RADIUS client setup

        :param servers: List or server settings to connect to
        :param attributes: Attributes to use
        :param retries: Integer, maximum retries before stopping to query the server
        :param timeout: Float, timeout value for connections
        """

        # Prepare dictionary
        self.dictionary = []

        attributes = kwargs.pop('attributes', None) or []
        for data in attributes:
            flags = data.get('flags')
            self.dictionary.append('ATTRIBUTE {name:20s} {oid:5d} {type:10s}{0}'.format(' {0}'.format(flags) if flags else '', **data))

        # Prepare clients settings
        self.servers = kwargs.pop('servers', None) or []

        if kwargs.get('retries'):
            self.retries = kwargs.pop('retries')

        if kwargs.get('timeout'):
            self.timeout = kwargs.pop('timeout')

        # Calling inherited
        return super().setup(*args, **kwargs)

    def validate_password(self, username, password, request):
        """ Attempt to send an authentication packet to servers """

        now = time.time()

        for settings in sorted(self.servers, key=lambda settings: settings.get('date', now)):
            kwargs = settings.copy()
            kwargs.pop('date', None)

            kwargs['dict'] = Dictionary(StringIO('\n'.join(self.dictionary)))

            client_ = Client(**kwargs)

            if self.retries:
                client_.retries = self.retries

            if self.timeout:
                client_.timeout = self.timeout

            req = client_.CreateAuthPacket(code=AccessRequest, username=username)
            req['password'] = req.PwCrypt(password)

            try:
                reply = client_.SendPacket(req)
            except Timeout:
                log.exception('Failed to communicate with RADIUS server (server=%s)', client_.server)
                continue

            client_._CloseSocket()    # pylint: disable=protected-access

            settings['date'] = now

            if reply.code == AccessAccept:
                return True

            # Invalid user
            log.error('[RADIUS] Authentication failed (server=%s, username=%s)', client_.server, username)

        return False


def parse_settings(settings, include_defaults=True):
    """ Parse and check RADIUS settings """

    if include_defaults:
        for regex in SETTINGS_DEFAULTS:
            found = False
            for section in settings:
                if not re.match(f'^{regex}$', section):
                    continue

                found = True

                for option, value in SETTINGS_DEFAULTS[regex].items():
                    settings[section].setdefault(option, value)

            if found or ':' in regex:
                continue

            settings[regex] = SETTINGS_DEFAULTS[regex].copy()

    for section in settings:
        settings[section].pop('here', None)

        for regex in SETTINGS_RULES:
            if re.match(f'^{regex}$', section):
                verify_rules(settings[section], SETTINGS_RULES[regex])
                break

    return settings


# Create client object
client = RadiusClient()


def includeme(config):
    """ RADIUS module initialization """

    # Load and parse settings
    settings = get_settings(config, 'radius')
    if settings is None:
        raise Exception('[RADIUS] Invalid or missing configuration for RADIUS, please check radius.filepath directive')

    settings = parse_settings(settings)

    # Settings
    attributes = []
    servers = []

    for section in settings:
        if section.startswith('attr:'):
            data = settings[section].copy()
            try:
                data['oid'] = int(section[5:])
            except ValueError as exc:
                raise Exception(f'[RADIUS] Invalid attribute definition for {section} in configuration') from exc

            attributes.append(data)

        elif section.startswith('server:'):
            data = settings[section].copy()
            if not data.pop('enabled'):
                continue

            data['secret'] = bytes(data['secret'], 'utf-8')

            servers.append(data)

    if not servers:
        raise Exception('[RADIUS] Empty or invalid configuration for server entries in configuration')

    # Prepare arguments
    kwargs = settings['radius'].copy()
    kwargs['attributes'] = attributes
    kwargs['servers'] = servers

    # Setup the client
    client.setup(**kwargs)

    log.info('[RADIUS] Initialization complete: servers=%s', ','.join(kwargs['server'] for kwargs in client.servers))



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" API REST interactive documentation view """

from collections import OrderedDict
import importlib
import inspect
import logging
import operator
import os
import re

from formencode.api import NoDefault

from pyramid.path import AssetResolver
from pyramid.renderers import render_to_response
from pyramid.security import NO_PERMISSION_REQUIRED

log = logging.getLogger(__name__)

RE_FUNC_PARAMS = re.compile(r'\((.*?)\)')
RE_LIBRARY_VERSION = re.compile(r"([A-Za-z0-9\.-]+)-(v?[0-9a-f\.]{2,}[0-9a-f]).*")


def datetime_format_python_to_moment(format_):
    """ Convert python datetime format (C Standard 1989) into Moment.js datetime format """

    # year 4 digits
    format_ = format_.replace('%Y', 'YYYY')
    # month 2 digits
    format_ = format_.replace('%m', 'MM')
    # day 2 digits
    format_ = format_.replace('%d', 'DD')
    # hours 2 digits (24h)
    format_ = format_.replace('%H', 'HH')
    # hours 2 digits (12h)
    format_ = format_.replace('%I', 'hh')
    # hours AM/PM
    format_ = format_.replace('%p', 'A')
    # minutes 2 digits
    format_ = format_.replace('%M', 'mm')
    # seconds 2 digits
    format_ = format_.replace('%S', 'ss')

    return format


def inspect_schema_doc(tree, parameters, parent=None):
    """ Inspect `formencode.Schema` documentation """

    if not parameters:
        return

    if isinstance(tree, list):
        for node in tree:
            inspect_schema_doc(node, parameters)
    else:
        if tree[0].__name__ in ('object', 'Declarative') or tree[0].__module__.startswith('formencode'):
            return

        doc = inspect.getdoc(tree[0])
        for line in doc.split('\n'):
            if line.startswith(':param'):
                line = line.split(':', 2)
                if len(line) != 3:
                    continue
                name = line[1].split(' ')[-1].strip()
                if parent is not None:
                    name = '{0}.{1}'.format(parent, name)
                if name in parameters:
                    parameters[name]['description'] = line[2].strip()


def inspect_validator(validator, name=None):
    """ Inspect `formencode.validators.FancyValidator` """

    data = dict(
        description=None,
    )

    if name is not None:
        data['name'] = name
        data['required'] = validator.not_empty is True
        if name in ('format', 'order', 'mode', 'page', 'page_size', 'paging', 'sort'):
            data['parameter_type'] = 'query special'
        else:
            data['parameter_type'] = 'query'

    if_empty = None
    if validator.if_empty is not None and validator.if_empty is not NoDefault:
        if_empty = validator.if_empty
    if_missing = None
    if validator.if_missing is not None and validator.if_missing is not NoDefault:
        if_missing = validator.if_missing
    data['default'] = if_missing or if_empty

    class_name = validator.__class__.__name__
    if class_name.endswith('Validator'):
        class_name = class_name.replace('Validator', '')

    data['type'] = class_name
    if class_name == 'OneOf':
        data['values'] = validator.list
    elif class_name in ('ForEach', 'NumberList', 'StringList'):
        data['items'] = inspect_validator(validator.validators[0]) if validator.validators else None
    elif class_name in ('Int', 'Number'):
        data['min'] = validator.min
        data['max'] = validator.max
    elif data['type'] == 'DateTime':
        if validator.is_date:
            data['type'] = 'Date'
        data['format'] = datetime_format_python_to_moment(validator.format)

    return data


# from CPython project
# https://hg.python.org/cpython/file/tip/Lib/inspect.py
def unwrap(func, stop=None):
    """Get the object wrapped by *func*.

    Follows the chain of :attr:`__wrapped__` attributes returning the last
    object in the chain.

    *stop* is an optional callback accepting an object in the wrapper chain
    as its sole argument that allows the unwrapping to be terminated early if
    the callback returns a true value. If the callback never returns a true
    value, the last object in the chain is returned as usual. For example,
    :func:`signature` uses this to stop unwrapping if any object in the
    chain has a ``__signature__`` attribute defined.

   :exc:`ValueError` is raised if a cycle is encountered.

    """
    if stop is None:
        def _is_wrapper(func):
            return hasattr(func, '__wrapped__')
    else:
        def _is_wrapper(func):
            return hasattr(func, '__wrapped__') and not stop(func)

    # remember the original func for error reporting
    func_ = func

    # Memoise by id to tolerate non-hashable objects
    memo = set([id(func_)])
    while _is_wrapper(func):
        func = func.__wrapped__
        id_func = id(func)
        if id_func in memo:
            raise ValueError('wrapper loop when unwrapping {!r}'.format(func_))
        memo.add(id_func)

    return func


# pylint: disable=too-many-branches,too-many-locals,too-many-statements
def api_doc(request, renderer_values, package):
    """Render API documentation

    Args:
        request (obj): currently active request.
        renderer_values (dict): values passed to renderer.
            breadcrumb: Page breadcrumb
            hide_undocumented: True to hide undocumented services (default False)
            title: Page title
            subtitle: Page subtitle
        package (string): requester package name.

    Returns:
        unicode: HTML response body.

    """

    introspector = request.registry.introspector
    services = OrderedDict()

    for item in introspector.get_category('routes'):
        introspectable = item['introspectable']
        discriminator = introspectable.discriminator

        if not discriminator.startswith('api.'):
            continue

        route_intr = introspector.get('routes', discriminator)
        route = route_intr['object']

        services[discriminator] = dict(
            allowed=True,
            doc=None,
            module=None,
            name=discriminator,
            pattern=request.script_name + route_intr['pattern'],
            parameters=[],
            request_methods=route_intr['request_methods'],
        )

        # get third-party predicates
        thrid_party_predicates = []
        for predicate in route.predicates:
            if predicate.__class__.__name__ == 'EnumPredicate':
                for param, values in predicate.params.items():
                    thrid_party_predicates.append(param)
                    services[discriminator]['parameters'].append(dict(
                        name=param,
                        type='OneOf',
                        parameter_type='path',
                        required=True,
                        default=None,
                        description=None,
                        values=values,
                    ))
            elif predicate.__class__.__name__ == 'NumericPredicate':
                for name in predicate.names:
                    thrid_party_predicates.append(name)
                    services[discriminator]['parameters'].append(dict(
                        name=name,
                        type='Number',
                        parameter_type='path',
                        required=True,
                        default=None,
                        description=None,
                        max=None,
                        min=None,
                    ))

        # get other predicates (not third-party) present in pattern
        for predicate_name in route.match(route_intr['pattern']).keys():
            if predicate_name not in thrid_party_predicates:
                services[discriminator]['parameters'].append(dict(
                    name=predicate_name,
                    type='ByteString',
                    parameter_type='path',
                    required=True,
                    default=None,
                    description=None,
                ))

    for item in introspector.get_category('permissions'):
        introspectable = item['introspectable']
        permission = introspectable.discriminator
        if permission == NO_PERMISSION_REQUIRED:
            continue

        for introspectable_view in item['related']:
            discriminator = introspectable_view['route_name']
            if discriminator in services:
                services[discriminator]['allowed'] = request.has_permission(permission).__class__.__name__ == 'ACLAllowed'

    for item in introspector.get_category('views'):
        introspectable = item['introspectable']
        discriminator = introspectable['route_name']

        if discriminator not in services:
            continue

        if introspectable['request_methods']:
            if isinstance(introspectable['request_methods'], tuple):
                services[discriminator]['request_methods'] = introspectable['request_methods']
            else:
                services[discriminator]['request_methods'] = (introspectable['request_methods'],)

        services[discriminator]['module'] = introspectable['callable'].__venusian_callbacks__['pyramid'][0][1]

        callable_func = unwrap(introspectable['callable'])
        source_lines = inspect.getsourcelines(callable_func)
        for line in source_lines[0]:
            if line.startswith('@validate'):
                parameters = dict()

                schema_name = re.findall(RE_FUNC_PARAMS, line)[0].split(',')[1].strip()
                source_path = inspect.getsourcefile(callable_func)
                with open(source_path, 'r') as fp:
                    for module_line in fp.readlines():
                        if module_line.startswith('from') and schema_name in module_line.split():
                            schema_module_name = module_line.split()[1]
                            break

                schema_module = importlib.import_module(schema_module_name)
                schema = getattr(schema_module, schema_name)

                for name, field in schema.fields.items():
                    class_name = field.__class__.__name__
                    if class_name.endswith('Schema'):
                        for sub_name, sub_field in field.__dict__['fields'].items():
                            parameter_name = '{0}.{1}'.format(name, sub_name)
                            parameters[parameter_name] = inspect_validator(sub_field, parameter_name)

                        inspect_schema_doc((field.__class__,), parameters, name)
                    else:
                        parameters[name] = inspect_validator(field, name)

                inspect_schema_doc(inspect.getclasstree(inspect.getmro(schema), unique=True), parameters)

                # sort parameters by parameter_type and name
                parameters = list(parameters.values())
                parameters.sort(key=operator.itemgetter('parameter_type', 'name'))

                services[discriminator]['parameters'] += parameters
                break

        doc = inspect.getdoc(callable_func)
        if doc is not None:
            doc %= request.registry.settings
        services[discriminator]['doc'] = doc

    modules = OrderedDict()
    orphans = []
    for discriminator, service in services.items():
        if service['module'] is None:
            # service has a route but no view!
            orphans.append(service)
            continue

        if service['module'] not in modules:
            module = importlib.import_module(service['module'])
            modules[service['module']] = dict(
                doc=module.__doc__.strip() if module.__doc__ else None,
                services=[],
            )
        modules[service['module']]['services'].append(service)
    renderer_values['modules'] = modules
    renderer_values['orphans'] = orphans

    # get versions of available JavaScript libraries
    renderer_values['static_lib_versions'] = dict()
    asset_resolver = AssetResolver(package)
    static_lib_path = asset_resolver.resolve('static/lib').abspath()
    for item in os.listdir(static_lib_path):
        matches = RE_LIBRARY_VERSION.findall(item)
        if len(matches) == 1:
            name, version = matches[0]
            renderer_values['static_lib_versions'][name] = version

    renderer_values['package'] = package

    # read CSS and JavaScript code, it must be inlined in HTML page
    bs_major_revision = renderer_values['static_lib_versions']['bootstrap'][0]
    asset_resolver = AssetResolver('pyramid_helpers')
    with open(asset_resolver.resolve('static/css/api-doc-bs{0}.css'.format(bs_major_revision)).abspath(), 'r') as fp:
        renderer_values['css'] = fp.read()
    with open(asset_resolver.resolve('static/js/api-doc-bs{0}.js'.format(bs_major_revision)).abspath(), 'r') as fp:
        renderer_values['js'] = fp.read()

    return render_to_response('/api-doc-bs{0}.mako'.format(bs_major_revision), renderer_values, request, package)



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Articles API
"""

from pyramid.view import view_config

from pyramid_helpers.forms import validate
from pyramid_helpers.forms.articles import CreateForm
from pyramid_helpers.forms.articles import SearchForm
from pyramid_helpers.forms.articles import StatusForm
from pyramid_helpers.funcs.articles import build_search_query
from pyramid_helpers.funcs.articles import create_or_modify
from pyramid_helpers.funcs.articles import get_article
from pyramid_helpers.models import DBSession
from pyramid_helpers.paginate import paginate


@view_config(route_name='api.articles.create', renderer='json', permission='articles.create')
@validate('article', CreateForm)
def create(request):
    """
    Creates an article
    """
    _ = request.translate
    form = request.forms['article']

    result = dict(
        method=request.path,
        apiVersion='1.0',
        params=form.params,
        result=True,
    )

    article = create_or_modify(request, form)
    if article is None:
        result['message'] = _('Missing or invalid parameter(s)')
        result['result'] = False
        result['errors'] = form.errors
        request.response.status_int = 400

        return result

    result['article'] = article.to_json()
    request.response.status_int = 201

    return result


@view_config(route_name='api.articles.delete', renderer='json', permission='articles.delete')
def delete(request):
    """
    Deletes an article
    """
    _ = request.translate

    result = dict(
        method=request.path,
        apiVersion='1.0',
        result=True,
    )

    article = get_article(request)
    if article is None:
        result['message'] = _('Invalid article')
        result['result'] = False
        request.response.status_int = 404

        return result

    DBSession.delete(article)

    return result


@view_config(route_name='api.articles.modify', renderer='json', permission='articles.modify')
@validate('article', CreateForm)
def modify(request):
    """
    Modifies an article
    """
    _ = request.translate
    form = request.forms['article']

    result = dict(
        method=request.path,
        apiVersion='1.0',
        params=form.params,
        result=True,
    )

    article = get_article(request)
    if article is None:
        result['message'] = _('Invalid article')
        result['result'] = False
        request.response.status_int = 404

        return result

    if not create_or_modify(request, form, article=article):
        result['message'] = _('Missing or invalid parameter(s)')
        result['result'] = False
        result['errors'] = form.errors
        request.response.status_int = 400

        return result

    result['article'] = article.to_json()

    return result


@view_config(route_name='api.articles.status', renderer='json', permission='articles.modify')
@validate('status', StatusForm)
def status(request):
    """
    Changes an article status
    """
    _ = request.translate
    form = request.forms['status']

    result = dict(
        method=request.path,
        apiVersion='1.0',
        params=form.params,
        result=True,
    )

    article = get_article(request)
    if article is None:
        result['message'] = _('Invalid article')
        result['result'] = False
        request.response.status_int = 404

        return result

    if form.errors:
        result['result'] = False
        result['errors'] = form.errors
        request.response.status_int = 400

        return result

    # Update object
    article.from_dict(form.result)

    result['article'] = article.to_json()

    return result


@view_config(route_name='api.articles.search', renderer='json', permission='articles.search')
@paginate('articles', limit=10, sort='id', order='desc')
@validate('search', SearchForm, method='get')
def search(request):
    """
    Searches in articles
    """
    form = request.forms['search']
    pager = request.pagers['articles']

    result = dict(
        method=request.path,
        apiVersion='1.0',
        params=form.params,
        result=True,
    )

    if form.errors:
        result['result'] = False
        result['errors'] = form.errors
        request.response.status_int = 400

        return result

    query = build_search_query(order=pager.order, sort=pager.sort, **form.result)
    pager.set_collection(query)

    result['articles'] = dict(
        items=[article.to_json(context='api.search') for article in pager],
        pager=dict(
            limit=pager.limit,
            page=pager.page,
            pages=pager.page_count,
            total=pager.item_count,
        )
    )

    return result


@view_config(route_name='api.articles.visual', renderer='json', permission='articles.visual')
def visual(request):
    """
    Gets an article
    """
    _ = request.translate

    result = dict(
        method=request.path,
        apiVersion='1.0',
        result=True,
    )

    article = get_article(request)
    if article is None:
        result['message'] = _('Invalid article')
        result['result'] = False
        request.response.status_int = 404

        return result

    result['article'] = article.to_json()

    return result



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Utils functions for Pyramid """

from configparser import ConfigParser
from configparser import Error
from functools import lru_cache
import importlib
import logging
import os
import random
import string

import pytz


RANDOM_STRING = string.ascii_letters + string.digits
TIMEZONE = 'Europe/Paris'


log = logging.getLogger(__name__)


def on_before_renderer(event):
    """ Add utc<->local datetime converters to context """

    request = event['request']

    event['localtoutc'] = request.localtoutc
    event['utctolocal'] = request.utctolocal


def on_new_request(event):
    """ Add utc<->local datetime converters to request """

    request = event.request

    def localtoutc(dt, request=request):
        tzinfo = get_tzinfo(request)
        offset = tzinfo.utcoffset(dt)
        return dt - offset

    request.localtoutc = localtoutc

    def utctolocal(dt, request=request):
        tzinfo = get_tzinfo(request)
        return tzinfo.fromutc(dt)

    request.utctolocal = utctolocal


def get_settings(obj, key, section=None):
    """
    Get settings from key
    If key is missing, try to load settings from `key.filepath` INI file
    """

    registry = obj.registry
    settings = registry.settings

    if key not in settings:
        # Load settings from file
        filepath = settings.get('{0}.filepath'.format(key))
        if filepath is None:
            log.debug('Missing %s.filepath directive, please check configuration file', key)
            return None

        filepath = os.path.abspath(filepath)
        if not os.path.isfile(filepath):
            log.error('Invalid %s.filepath directive, please check configuration file', key)
            return None

        if not os.access(filepath, os.R_OK):
            log.error('Not enough permission to access file %s', filepath)
            return None

        parser = ConfigParser(defaults=dict(here=os.path.dirname(filepath)))
        parser.optionxform = str    # Be case sensitive

        try:
            with open(filepath, 'r') as fp:
                parser.read_file(fp)

            settings[key] = dict(
                (section, dict(parser.items(section)))
                for section in parser.sections()
            )
        except (IOError, Error):
            log.exception('Failed to read file %s', filepath)
            return None

    if section is None:
        return settings[key]

    return settings[key].get(section) or {}


@lru_cache()
def get_tzinfo(request):
    """ Get timezone object """

    # Try to get timezone from authenticated user
    if getattr(request, 'authenticated_user', None) is not None:
        timezone = getattr(request.authenticated_user, 'timezone', None)
    else:
        timezone = None

    if timezone is None or timezone not in pytz.common_timezones:
        # Falling back to timezone from settings
        registry = request.registry
        settings = registry.settings

        timezone = settings.get('timezone')

    if timezone is None or timezone not in pytz.common_timezones:
        # Use default timezone
        timezone = TIMEZONE

    return pytz.timezone(timezone)


def random_string(length=10):
    """ Get a random string """

    return ''.join(
        random.choice(RANDOM_STRING)
        for _ in range(length)
    )


def resolve_dotted(dotted):
    """ Resolve dotted string to module attribute """

    if not dotted:
        return None

    module_name, module_attr = dotted.rsplit('.', 1)
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        return None

    return getattr(module, module_attr)


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.utils')
    """

    registry = config.registry
    settings = registry.settings

    # Subscribers setup
    config.add_subscriber(on_before_renderer, 'pyramid.events.BeforeRender')
    config.add_subscriber(on_new_request, 'pyramid.events.NewRequest')

    # LDAP configuration
    if settings.get('ldap.enabled', 'false').lower() in ('true', 'yes'):
        config.include('pyramid_helpers.ldap')

    # RADIUS configuration
    if settings.get('radius.enabled', 'false').lower() in ('true', 'yes'):
        config.include('pyramid_helpers.radius')



# -*- coding: utf-8 -*-

# pyramid-helpers -- Helpers to develop Pyramid applications
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#     Valéry Febvre <vfebvre@easter-eggs.com>
#
# Copyright (C) 2011-2021 Cyril Lacoux, Easter-eggs
# https://gitlab.com/yack/pyramid-helpers
#
# This file is part of pyramid-helpers.
#
# pyramid-helpers is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# pyramid-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Authentication module for Pyramid-Helpers

Useful documentation can be found at:
 * https://docs.pylonsproject.org/projects/pyramid/en/latest/api/security.html
 * https://docs.pylonsproject.org/projects/pyramid-cookbook/en/latest/auth/index.html
"""

import hashlib
import logging

from beaker.util import NoneType
from beaker.util import verify_rules

from pyramid.authentication import AuthTktCookieHelper
from pyramid.authentication import CallbackAuthenticationPolicy
from pyramid.authentication import extract_http_basic_credentials
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.interfaces import IAuthenticationPolicy

from zope.interface import implementer

from pyramid_helpers.utils import get_settings
from pyramid_helpers.utils import resolve_dotted


AUTH_CLIENTS = {}

SETTINGS_DEFAULTS = {
    'auth': {
        'policies': ['ticket', ],
        'mode': None,
        'callback': 'pyramid_helpers.auth.authentication_callback',
        'get_user_by_username': None,
    },
    'policy:basic': {
        'realm': 'Pyramid-Helpers Application',
    },
    'policy:remote': {
        'fake_user': None,
        'login_url': None,
        'logout_url': None,
    },
    'policy:ticket': {
        'secret': 'the-big-secret-for-secured-authentication',
        'hashalg': 'sha512',
    },
    'policy:token': {
        'header': 'X-PH-Authentication-Token',
        'query_param': None,
        'get_username_by_token': None,
    },
}

SETTINGS_RULES = {
    'auth': [
        ('enabled', (bool, NoneType), 'enabled must be a boolean or an integer'),
        ('policies', (list, NoneType), 'policies must be a comma separated list of policies'),
        ('mode', (str, NoneType), 'mode must be a string designating a valid mode'),
        ('callback', (str, NoneType), 'callback must be a string designating a valid callback'),
        ('get_user_by_username', (str,), 'get_user_by_username must be a string designating a valid callback'),
    ],
    'policy:basic': [
        ('realm', (str, NoneType), 'realm must be a string designating a realm authentication'),
    ],
    'policy:remote': [
        ('fake_user', (str, NoneType), 'fake_user must be a string designating a username'),
        ('login_url', (str, NoneType), 'login_url must be a string designating a valid url'),
        ('logout_url', (str, NoneType), 'logout_url must be a string designating a valid url'),
    ],
    'policy:ticket': [
        ('secret', (str, NoneType), 'secret must be a string designating a secret'),
        ('hashalg', (str, NoneType), 'hashalg must be a string designating a valid hashalg'),
    ],
    'policy:token': [
        ('header', (str, NoneType), 'header must be a string designating a HTTP header'),
        ('query_param', (str, NoneType), 'query_param must be a string designating a query parameter'),
        ('get_username_by_token', (str, NoneType), 'get_user_by_token must be a string designating a valid callback'),
    ],
}


log = logging.getLogger(__name__)


class AuthenticationClient:
    """ Authentication client base class """

    __auth_mode__ = None

    def __enter__(self):
        """
        This method is called by the `with:` statement and should be overriden
        to execute some tasks **before** validating the password.
        """

        return self

    def __exit__(self, type_, value, traceback):
        """
        This method is called by the `with:` statement and should be overriden
        to execute some tasks **after** validating the password.
        """

    # pylint: disable=unused-argument
    def setup(self, *args, **kwargs):
        """ Register authentication client """

        if self.__auth_mode__ is None:
            log.error('[AUTH] Attribute `.__auth_mode__` of AuthenticationClient instance must be set')
            return False

        if self.__auth_mode__ in AUTH_CLIENTS:
            log.error('[AUTH] An authentication client is already registered for mode %s', self.__auth_mode__)
            return False

        # Registering object
        AUTH_CLIENTS[self.__auth_mode__] = self

        log.info('[AUTH] Registered authentication client, mode=%s, client=%s', self.__auth_mode__, self.__class__.__name__)

        return True

    def validate_password(self, username, password, request):
        """ Please, override this method to implement the password validation """

        raise NotImplementedError()


@implementer(IAuthenticationPolicy)
class MultiAuthenticationPolicy(CallbackAuthenticationPolicy):
    """
    A Pyramid authentication policy that implements the following protocols:
     * ticket
     * remote
     * basic
     * token
    """

    POLICIES = ['ticket', 'remote', 'basic', 'token']

    # pylint: disable=too-many-arguments
    def __init__(self, callback, policies, hashalg='sha512', realm='Pyramid Helpers', secret=None):

        for policy in policies:
            assert policy in self.POLICIES, '[AUTH] Invalid policy: {0}'.format(policy)

        self.cookie = AuthTktCookieHelper(secret, hashalg=hashalg) if 'ticket' in policies else None
        self.policies = policies
        self.realm = realm

        self.__callback = callback

    def callback(self, username, request):
        """ Custom method that handles multiple authentication mechanisms """

        if request.authentication_policy == 'basic':
            credentials = extract_http_basic_credentials(request)
            if credentials is None:
                return None

            if not check_credentials(credentials.username, credentials.password, request):
                return None

        return self.__callback(username, request)

    def forget(self, request):
        """ Custom method that handles multiple authentication mechanisms """

        if request.authentication_policy == 'basic':
            # Ask client to re-authenticate
            return [('WWW-Authenticate', 'Basic realm="%s"' % self.realm)]

        if self.cookie:
            # Delete cookie
            return self.cookie.forget(request)

        return []

    def remember(self, request, username, **kw):
        """
        Custom method that handles multiple authentication mechanisms

        Accepts the following kw args: ``max_age=<int-seconds>, ``tokens=<sequence-of-ascii-strings>``.
        """

        if request.authentication_policy != 'ticket':
            # Do nothing
            return []

        # Set cookie
        return self.cookie.remember(request, username, **kw)

    def unauthenticated_userid(self, request):
        """ Custom method that handles multiple authentication mechanisms """

        # Extract data
        credentials = extract_http_basic_credentials(request)
        policy = request.headers.get('X-PH-Authentication-Policy')
        remote_user = request.environ.get('REMOTE_USER')
        token = extract_token(request)

        # Guess and check policy
        if policy is None:
            if remote_user:
                policy = 'remote'

            elif credentials:
                policy = 'basic'

            elif token:
                policy = 'token'

            elif self.cookie:
                policy = 'ticket'

        if policy not in self.policies:
            policy = self.policies[0]

        # Store policy to request
        # Policy may be set by the fake user tween
        if getattr(request, 'authentication_policy', None) is None:
            request.authentication_policy = policy

        if policy == 'basic':
            if credentials is None:
                return None

            return credentials.username

        if policy == 'remote':
            return remote_user

        if policy == 'token':
            return get_username_by_token(token, request)

        # Ticket
        result = self.cookie.identify(request)
        if result:
            return result['userid']

        return None


# pylint: disable=unused-argument
def auth_fake_user_tween_factory(handler, registry):
    """
    Tween that adds a fake user to environ to simulate remote user based
    authentication.
    """

    def auth_fake_user_tween(request):
        # Get fake user from settings
        registry = request.registry
        settings = registry.settings
        username = settings['auth']['policy:remote']['fake_user']

        # Add fake user as REMOTE_USER if requested
        environ = request.environ
        environ['REMOTE_USER'] = username

        # Set the authentication policy
        request.authentication_policy = 'fake_user'

        return handler(request)

    return auth_fake_user_tween


def authentication_callback(username, request):
    """ Called by authentication policy when user is authenticated """

    user = get_user_by_username(username, request)
    if user is None:
        # Invalid user
        return None

    return [
        ('group:{0}'.format(group.name))
        for group in user.groups
    ]


def check_credentials(username, password, request):
    """ Check extracted credential using configured mode """

    params = get_settings(request, 'auth', 'auth')

    auth_client = AUTH_CLIENTS.get(params['mode'])
    if auth_client is None:
        # Invalid mode
        return False

    with auth_client as client:
        return client.validate_password(username, password, request)


def extract_token(request):
    """ Extract authentication token from request """

    params = get_settings(request, 'auth', 'policy:token')

    if params.get('query_param'):
        return request.GET.get(params['query_param'])

    if params.get('header'):
        return request.headers.get(params['header'])

    return None


def get_user_by_username(username, request):
    """ Wrapper to get_user_by_username function defined in settings """

    params = get_settings(request, 'auth', 'auth')
    func = params['get_user_by_username']

    return func(username, request)


def get_username_by_token(token, request):
    """ Wrapper to get_username_by_token function defined in settings """

    params = get_settings(request, 'auth', 'auth')
    func = params['get_username_by_token']

    return func(token, request)


def on_before_renderer(event):
    """ Add authenticated_user and has_permission() to renderer context """

    request = event['request']

    event['authentication_policy'] = request.authentication_policy
    event['authenticated_user'] = request.authenticated_user
    event['has_permission'] = request.has_permission


def on_new_request(event):
    """ Add authenticated_user to request """

    request = event.request
    request.authenticated_user = get_user_by_username(request.authenticated_userid, request)

    if not hasattr(request, 'authentication_policy'):
        request.authentication_policy = None


def parse_settings(settings, include_defaults=True):
    """ Parse and check authentication settings """

    if include_defaults:
        for section in SETTINGS_DEFAULTS:
            if section not in settings:
                settings[section] = SETTINGS_DEFAULTS[section].copy()
                continue

            for option, value in SETTINGS_DEFAULTS[section].items():
                settings[section].setdefault(option, value)

    for section, rules in SETTINGS_RULES.items():
        if section in settings:
            verify_rules(settings[section], rules)

    # Check policies
    if 'all' in settings['auth']['policies']:
        settings['auth']['policies'] = MultiAuthenticationPolicy.POLICIES[:]
    else:
        for policy in settings['auth']['policies']:
            if policy not in MultiAuthenticationPolicy.POLICIES:
                raise Exception('[AUTH] Invalid policy for parameters policies in section [auth]: {0}'.format(policy))

    # Check mode
    if settings['auth']['mode'] is None:
        raise Exception('[AUTH] Missing value for parameter mode in section [auth]')

    if settings['auth']['mode'] not in AUTH_CLIENTS:
        raise Exception('[AUTH] Invalid value for parameter mode in section [auth]: {0}'.format(settings['auth']['mode']))

    # Check hashalg
    if settings['policy:ticket']['hashalg'] not in hashlib.algorithms_available:
        raise Exception('[AUTH] Invalid value for parameter hashalg in section [policy:ticket]: {0}'.format(settings['auth']['policy:ticket']['hashalg']))

    # Resolve authentication callback
    func = resolve_dotted(settings['auth']['callback'])
    if func is None:
        raise Exception('[AUTH] Invalid callback: {0}'.format(settings['auth']['callback']))
    settings['auth']['callback'] = func

    # Resolve get_user_by_username
    func = resolve_dotted(settings['auth']['get_user_by_username'])
    if func is None:
        raise Exception('[AUTH] Invalid value for parameter get_user_by_username in section [auth]: {0}'.format(settings['auth']['get_user_by_username']))
    settings['auth']['get_user_by_username'] = func

    # Resolve get_username_by_token if needed
    if 'token' in settings['auth']['policies']:
        func = resolve_dotted(settings['policy:token']['get_username_by_token'])
        if func is None:
            raise Exception('[AUTH] Invalid value for parameter get_username_by_token in section [policy:token]: {0}'.format(settings['auth']['get_username_by_token']))
        settings['policy:token']['get_username_by_token'] = func

    return settings


def includeme(config):
    """
    Set up standard configurator registrations. Use via:

    .. code-block:: python

    config = Configurator()
    config.include('pyramid_helpers.auth')
    """

    config.add_subscriber(on_before_renderer, 'pyramid.events.BeforeRender')
    config.add_subscriber(on_new_request, 'pyramid.events.NewRequest')

    # Load an parse settings
    settings = get_settings(config, 'auth')
    if settings is None:
        raise Exception('[AUTH] Invalid or missing configuration for AUTH, please check auth.filepath directive')

    settings = parse_settings(settings)

    # Add fake user tween if needed
    if 'remote' in settings['auth']['policies'] and settings['policy:remote']['fake_user'] is not None:
        log.warning('[AUTH] POLICY `remote` IS ENABLED WITH `fake_user` SET TO `%s`, USER `%s` WILL BE AUTOMATICALLY CONNECTED WITHOUT ANY AUTHENTICATION, THIS IS VERY DANGEROUS !!!', settings['policy:remote']['fake_user'], settings['policy:remote']['fake_user'])

        config.add_tween('pyramid_helpers.auth.auth_fake_user_tween_factory')

    # Set authentication policy
    config.set_authentication_policy(MultiAuthenticationPolicy(
        settings['auth']['callback'],
        settings['auth']['policies'],
        hashalg=settings['policy:ticket']['hashalg'],
        realm=settings['policy:basic']['realm'],
        secret=settings['policy:ticket']['secret'],
    ))

    # Set authorization policy
    config.set_authorization_policy(ACLAuthorizationPolicy())

    log.info('[AUTH] Initialization complete: policies=%s, mode=%s', settings['auth']['policies'], settings['auth']['mode'])
